#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <linux/cdrom.h>
#include <fcntl.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <string.h>
#include <time.h>

#include "CDVD.h"
#include "interface.h"
#include "support.h"

union {
	struct cdrom_msf msf;
	unsigned char buf[CD_FRAMESIZE_RAW];
} cr;

static inline int msf_to_lba(char m, char s, char f) {
	return (((m * CD_SECS) + s) * CD_FRAMES + f) - CD_MSF_OFFSET;
}

u8 *cdbuffer;

int cdHandle = -1;
int _Thread = 0;
pthread_t thread = -1;
int readed, subqread;
s32 disctype;
s32 traystatus;

void *CdrThread(void *arg);

#define CreateThread() { \
	if (_Thread == 0) { _Thread = 1; pthread_create(&thread, NULL, CdrThread, NULL); } }

#define ExitThread() { \
	{ _Thread = 0; pthread_exit(NULL); } }

const unsigned char version  = PS2E_CDVD_VERSION;
const unsigned char revision = 0;
const unsigned char build    = 3;    // increase that with each version

static char *libraryName      = "CDVDlinuz Driver";

u32 CALLBACK PS2EgetLibType() {
	return PS2E_LT_CDVD;
}

char* CALLBACK PS2EgetLibName() {
	return libraryName;
}

u32 CALLBACK PS2EgetLibVersion2(u32 type) {
	return (version<<16)|(revision<<8)|build;
}

GtkWidget *MsgDlg;

void OnMsg_Ok() {
	gtk_widget_destroy(MsgDlg);
	gtk_main_quit();
}

void SysMessage(char *fmt, ...) {
	GtkWidget *Ok,*Txt;
	GtkWidget *Box,*Box1;
	va_list list;
	char msg[512];

	va_start(list, fmt);
	vsprintf(msg, fmt, list);
	va_end(list);

	if (msg[strlen(msg)-1] == '\n') msg[strlen(msg)-1] = 0;

	MsgDlg = gtk_window_new (GTK_WINDOW_DIALOG);
	gtk_window_set_position(GTK_WINDOW(MsgDlg), GTK_WIN_POS_CENTER);
	gtk_window_set_title(GTK_WINDOW(MsgDlg), "CDVDlinuz Msg");
	gtk_container_set_border_width(GTK_CONTAINER(MsgDlg), 5);

	Box = gtk_vbox_new(5, 0);
	gtk_container_add(GTK_CONTAINER(MsgDlg), Box);
	gtk_widget_show(Box);

	Txt = gtk_label_new(msg);
	
	gtk_box_pack_start(GTK_BOX(Box), Txt, FALSE, FALSE, 5);
	gtk_widget_show(Txt);

	Box1 = gtk_hbutton_box_new();
	gtk_box_pack_start(GTK_BOX(Box), Box1, FALSE, FALSE, 0);
	gtk_widget_show(Box1);

	Ok = gtk_button_new_with_label("Ok");
	gtk_signal_connect (GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnMsg_Ok), NULL);
	gtk_container_add(GTK_CONTAINER(Box1), Ok);
	GTK_WIDGET_SET_FLAGS(Ok, GTK_CAN_DEFAULT);
	gtk_widget_show(Ok);

	gtk_widget_show(MsgDlg);	

	gtk_main();
}

s32 CALLBACK CDVDinit() {
	return 0;
}

s32 CALLBACK CDVDopen() {
	LoadConf();

	if (cdHandle > 0) return 0;	/* it's already open */
	cdHandle = open(conf.Dev, O_RDONLY);
	if (cdHandle != -1) { // if we can't open the cdrom we'll works as a null plugin
		ioctl(cdHandle, CDROM_LOCKDOOR, NULL);
		ioctl(cdHandle, CDROMSTART, NULL);
	} else {
		printf("CDR: Could not open %s, working as a null plugin\n", conf.Dev);
	}

	return 0;
}

void CALLBACK CDVDclose() {
	if (cdHandle < 1) return;
//	CDVDstop();
	close(cdHandle);
	cdHandle = -1;
/*	if (thread != -1) {
		pthread_cancel(thread);
		thread = -1;
	}*/
}

void CALLBACK CDVDshutdown() {
}

s32  CALLBACK CDVDreadTrack(u32 lsn, int mode) {
	if (cdHandle < 1) {
		memset(cr.buf, 0, DATA_SIZE);
		return 0;
	}

//	printf ("CDVDreadTrack %d, mode %d\n", lsn, mode);

	cr.msf.cdmsf_min0 = (lsn / 75) / 60;
	cr.msf.cdmsf_sec0 = (lsn / 75) % 60 + 2;
	cr.msf.cdmsf_frame0 = lsn % 75;

//	if (ReadMode == NORMAL) {
		if (ioctl(cdHandle, CDROMREADRAW, &cr) == -1) {
			printf("error reading from cdrom %d\n", lsn);
			return -1;
		}
/*	} else {					// THREADED
		readed = 0xff;
		CreateThread();
	}
	subqread = 0;*/

	switch (mode) {
		case CDVD_MODE_2352: cdbuffer = cr.buf;    break;
		case CDVD_MODE_2340: cdbuffer = cr.buf+12; break;
		case CDVD_MODE_2328: cdbuffer = cr.buf+24; break;
		case CDVD_MODE_2048: cdbuffer = cr.buf+24; break;
	}

	return 0;
}

u8*  CALLBACK CDVDgetBuffer() {
/*	if (ReadMode == THREADED) {
		while (readed == 0xff) {}
		pthread_join(thread, NULL);
		if (readed == -1)
			return NULL;
	}
*/
	return cdbuffer;
}
/*
void *CdrThread(void *arg) {
	readed = ioctl(cdHandle, CDROMREADRAW, &cr);
	ExitThread();
}*/

s32  CDVDgetTN(cdvdTN *Buffer) {
	struct cdrom_tochdr toc;

	if (cdHandle < 1) {
		Buffer->strack = 1;
		Buffer->etrack = 1;
		return 0;
	}

	if (ioctl(cdHandle, CDROMREADTOCHDR, &toc) == -1)
		return -1;

	Buffer->strack = toc.cdth_trk0;	// start track
	Buffer->etrack = toc.cdth_trk1;	// end track

	return 0;
}

s32  CALLBACK CDVDgetTD(u8 Track, cdvdTD *Buffer) {	//track info: min,sec,frame,type
	struct cdrom_tocentry entry;

	if (cdHandle < 1) {
		return 0;
	}

	if (Track == 0)
		Track = 0xaa;			// total time
	entry.cdte_track = Track;
	entry.cdte_format = CDROM_MSF;

	if (ioctl(cdHandle, CDROMREADTOCENTRY, &entry) == -1)
		return -1;

	Buffer->minute = entry.cdte_addr.msf.minute;
	Buffer->second = entry.cdte_addr.msf.second;
	Buffer->frame  = entry.cdte_addr.msf.frame;
	Buffer->type   = entry.cdte_datamode;

	return 0;
}

s32  CALLBACK CDVDgetType() {
	static time_t to;
	int ret;

	if (cdHandle < 1) return 0;
	if (to == time(NULL)) return disctype;

	ret = ioctl(cdHandle, CDROM_DRIVE_STATUS);
	if (ret == CDS_TRAY_OPEN || ret == CDS_NO_DISC || ret == -1) {
		disctype = CDVD_TYPE_NODISC;
	} else {
		ret = ioctl(cdHandle, CDROM_DISC_STATUS);
		if (ret == CDS_TRAY_OPEN || ret == CDS_NO_DISC) disctype = CDVD_TYPE_NODISC;

		if (ret == CDS_AUDIO) disctype = CDVD_TYPE_CDDA; else
		if (ret == CDS_MIXED) disctype = CDVD_TYPE_PS2CDDA;
		else disctype = CDVD_TYPE_PS2CD;
	}

	to = time(NULL);
	return disctype;
}

s32  CALLBACK CDVDgetTrayStatus() {
	static time_t to;
	int ret;

	if (to == time(NULL)) return traystatus;

	ret = ioctl(cdHandle, CDROM_DRIVE_STATUS);
	if (ret == CDS_TRAY_OPEN) traystatus = CDVD_TRAY_OPEN;
	else traystatus = CDVD_TRAY_CLOSE;

	to = time(NULL);
	return traystatus;
}

GtkWidget *ConfDlg;
GtkWidget *Edit;
GtkWidget *Combo;
GList *glist;

void OnConf_Ok() {
	char *tmp;
	int i;

	tmp = gtk_entry_get_text(GTK_ENTRY(Edit));
	strcpy(conf.Dev, tmp);
	tmp = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(Combo)->entry));
	for (i = 0; i < READ_MODES; i++)
		if (!strcmp(ReadModeNames[i], tmp))
			conf.ReadMode = i;
	gtk_widget_destroy(ConfDlg);
	SaveConf();
	gtk_main_quit();
}

void OnConf_Cancel() {
	gtk_widget_destroy(ConfDlg);
	gtk_main_quit();
}

void CALLBACK CDVDconfigure() {
	int i;

	LoadConf();

	ConfDlg = create_Config();

	Combo = lookup_widget(ConfDlg, "GtkCombo_Mode");
	Edit  = lookup_widget(ConfDlg, "GtkEntry_Dev");

	gtk_entry_set_text(GTK_ENTRY(Edit), conf.Dev);

	glist = NULL;
	for (i = 0; i < READ_MODES; i++)
		glist = g_list_append(glist, ReadModeNames[i]);
	gtk_combo_set_popdown_strings(GTK_COMBO(Combo), glist);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(Combo)->entry), ReadModeNames[conf.ReadMode]);

	gtk_widget_show_all(ConfDlg);
	gtk_main();
}

GtkWidget *About;

void OnAbout_Ok() {
	gtk_widget_destroy(About);
	gtk_main_quit();
}

void CALLBACK CDVDabout() {
	About = create_About();

	gtk_widget_show_all(About);
	gtk_main();
}

s32 CALLBACK CDVDtest() {
	cdHandle = open(conf.Dev, O_RDONLY);
	if (cdHandle == -1)
		return -1;
	close(cdHandle);
	cdHandle = -1;
	return 0;
}

void LoadConf() {
	FILE *f;
	char cfg[255];

	sprintf(cfg, "%s/.PS2E/CDVDlinuz.cfg", getenv("HOME"));
	f = fopen(cfg, "r");
	if (f == NULL) {
		strcpy(conf.Dev, DEV_DEF);
		conf.ReadMode = NORMAL;
		return;
	}
	fscanf(f, "Dev = %s\n", conf.Dev);
	fscanf(f, "ReadMode = %d\n", &conf.ReadMode);
	fclose(f);
}

void SaveConf() {
	FILE *f;
	char cfg[255];

    sprintf (cfg, "%s/.PS2E", getenv("HOME"));
	mkdir(cfg, 0755);
	sprintf(cfg, "%s/.PS2E/CDVDlinuz.cfg", getenv("HOME"));
	f = fopen(cfg, "w");
	if (f == NULL)
		return;
	fprintf(f, "Dev = %s\n", conf.Dev);
	fprintf(f, "ReadMode = %d\n", conf.ReadMode);
	fclose(f);
}
